#include "iterfuncOrd.h"

using namespace std ;
	
namespace ord {


   	class AdmisNormalElement : public IterFuncNormalElement {
        static const Ordinal& theAdmisClassLimit ;
	protected:

        class AdmisLevParameters : public IterFuncParameters {
		public:
			const OrdinalImpl * leastSigParam;
			const Ordinal& indexCK ;
			const Ordinal& drillDown ;
            const Embeddings& embeddings ;
			AdmisLevParameters(const Ordinal& ixCK, const Ordinal& iter,
				const Ordinal * const * const params=NULL,
				const Ordinal& dd = Ordinal::zero,
                const Embeddings & embed = Embeddings::embedNone,
				int level = admisCodeLevel);
			virtual ~AdmisLevParameters(){}

		};


        bool validAdmisNormalElementInt() const;
	public:
        static bool validAdmisNormalElement(const Ordinal& inx,
            const Ordinal& iter, const Ordinal * const * const params,
            const Ordinal& dd, const Embeddings& embed=Embeddings::embedNone);

		const OrdinalImpl& indexCK ;
		const OrdinalImpl& indexCKmo ;
		const Ordinal& indexCKo ;
		const Ordinal& drillDown ;
		const OrdinalImpl & leastSigParam ;

		virtual const OrdinalImpl& getIndexCK() const {return indexCK;}

private :    
        const AdmisLevOrdinal * admissibleBase;
        const AdmisNormalElement& createAdmissibleBase() const ;
        const OrdinalImpl& kappaLim(const OrdinalImpl& effEmbIx) const;
public:
        int entryCount ;
        int entryInc() const
            {return ++(((AdmisNormalElement *) this)->entryCount);}
        int entryDec() const
            {return --(((AdmisNormalElement *) this)->entryCount);}
        const AdmisNormalElement & admissibleBaseElement ;
		
        const AdmisLevOrdinal& getAdmissibleBase() const ;
        const class AdmisLevOrdinalImpl& getAdmissibleBaseImpl() const ;
        const OrdinalImpl& getCompareIndexCK(
            const Embeddings* emb=NULL) const ;
        const OrdinalImpl& ixMone ;
        const Embeddings embeddings ;


		AdmisNormalElement(const Ordinal& indexCK, const Ordinal& iter,
			const Ordinal* const * const params, 
			const Ordinal& drillDown = Ordinal::zero,
            const Embeddings&embed = Embeddings::embedNone,
            Int fac=1);

		AdmisNormalElement(const AdmisLevParameters* prma, Int fac=1);

		static const int admisCodeLevel = iterFuncCodeLevel + 1 ;

        const Ordinal & leUse(Int n) const;
        const Ordinal & loUseLe(const OrdinalImpl& le) const;
        const Ordinal & loUse(const OrdinalImpl& ord) const;

        const AdmisNormalElement& addEmbeddings(const Embeddings& emb) const;


		virtual void normalFormName(string& base) const ;
		virtual void texNormalForm(string& str) const ;
        virtual void psiNormalForm(string& str) const ;
        const Ordinal& classLimit() const {return theAdmisClassLimit;}

        static const OrdinalImpl& indexCKtoLimitType(const OrdinalImpl&ix,
            const Ordinal&dd = Ordinal::zero);

		const AdmisNormalElement& addParam(const Ordinal& param) const;
		

		

		virtual const CantorNormalElement& getCopy(const Int fac=1) const;

		const OrdinalImpl& typeIndexCK() const ;
		const OrdinalImpl& limitInfo(CantorNormalElement::LimitTypeInfo& info)
			const ;
		
		CantorNormalElement::LimitTypeInfo getInfo() const {
			if (theLimitTypeInfo != unknownLimit) return theLimitTypeInfo ;
			CantorNormalElement::LimitTypeInfo ret ;
			limitInfo(ret); return ret ; }

        static const OrdinalImpl& boundedLimitType(const Ordinal& bound,
            const OrdinalImpl&type, const Ordinal& dd = Ordinal::zero);

		virtual const OrdinalImpl& limitType() const ;
		virtual const OrdinalImpl& maxLimitType() const ;

		virtual const OrdinalImpl& limitType(const Ordinal& context) const ;
		virtual const OrdinalImpl& maxLimitType(const Ordinal& context) const ;

        virtual const CantorNormalElement& limitMaxEmbed(
            const OrdinalImpl&lim) const ;

		virtual const OrdinalImpl& limitElement(Int n) const;
		virtual const OrdinalImpl& drillDownLimitElement(Int n) const;
        virtual const OrdinalImpl& embedLimitElement(Int n) const  ;

		const OrdinalImpl* limitOrdCom(const OrdinalImpl & ord) const;
		virtual const OrdinalImpl& limitOrd(const OrdinalImpl & ord) const ;
		virtual const CantorNormalElement & addFactors(
			const CantorNormalElement& toAdd) const ;

        const Embeddings& effectiveEmbedding(const Embeddings& termEmbd) const ;

        const OrdinalImpl& effectiveIndexCK( const Embeddings& embed) const ;

        virtual int compare(const Embeddings& embed,
            const Embeddings& termEmbed,
            const CantorNormalElement& trm, bool ignoreFactor = false) const ;

        int compare(const OrdinalImpl& trm) const
        {
            return CantorNormalElement::compare(trm);
        }


        

		virtual int compare(const CantorNormalElement& trm,
			bool ignoreFactor = false) const ;
		

		protected:
		const OrdinalImpl& dRep1(const Ordinal& nval) const;

		private:
        static const Ordinal & drillDownLimitValue ;

		virtual const OrdinalImpl & createVirtualOrdImpl(const Ordinal & ix,
			const Ordinal & lev, const Ordinal * const * const parames,
			const Ordinal& drillDownParams,
            const Embeddings& embed) const ;
		virtual const Ordinal & createVirtualOrd(const Ordinal & ix,
            const Ordinal & lev, const Ordinal * const * const params,
			const Ordinal& drillDownParams, const Embeddings& embed) const ;

        virtual const OrdinalImpl & createVirtualOrdImpl(const Ordinal & lev,
			const Ordinal * const * const parames=NULL) const ;
		virtual const Ordinal & createVirtualOrd(const Ordinal & lev,
			const Ordinal * const * const parames=NULL) const ;



		virtual const OrdinalImpl & createVirtualOrdImpl(
			const Ordinal * const * const parames) const ;

		virtual const Ordinal & createVirtualOrd(
			const Ordinal * const * const parames) const ;


	

	};


	class AdmisLevOrdinalImpl: public IterFuncOrdinalImpl {
		friend class AdmisNormalElement ;
		friend class AdmisLevOrdinal ;
		static const NormalFormTerm& createTerms(const Ordinal& indexCK,
			const Ordinal& iter, const Ordinal* const * const params,
			const Ordinal& drillDown = Ordinal::zero,
            const Embeddings& embed=Embeddings::embedNone);
        

		static string makeName(const Ordinal& ixCK, const Ordinal& iter,
			const Ordinal* const * const params,
			const Ordinal& drillDown = Ordinal::zero,
            const Embeddings& embed=Embeddings::embedNone);
		static string makeTexName(const Ordinal& ixCK, const Ordinal& iter,
			const Ordinal* const * const params,
			const Ordinal& drillDown = Ordinal::zero,
            const Embeddings& embed=Embeddings::embedNone);
    public:

		AdmisLevOrdinalImpl(const Ordinal& indexCK, const Ordinal& iter,
			const Ordinal* const * const params=NULL,
			const Ordinal& drillDown = Ordinal::zero,
            const Embeddings& embed=Embeddings::embedNone):
			IterFuncOrdinalImpl(makeName(indexCK,iter, params,drillDown,embed),
					createTerms(indexCK,iter, params,drillDown,embed)){}

		AdmisLevOrdinalImpl(const AdmisNormalElement& elt):
			IterFuncOrdinalImpl(elt){}

		AdmisLevOrdinalImpl(const string& name,const NormalFormTerm&  trms):
			IterFuncOrdinalImpl(name,trms){}

        const AdmisLevOrdinal* getAdmissibleBase() const ;

			
		virtual ~AdmisLevOrdinalImpl(){}

        const AdmisNormalElement * getFirstAdmis() const ;
        
		const OrdinalImpl& getIndexCK() const ;
        
        

        const OrdinalImpl& addEmbeddings(const Embeddings &embed) const;

	};

	class AdmisLevOrdinal : public IterFuncOrdinal {
	public:
		AdmisLevOrdinal (const Ordinal& indexCK, const Ordinal& ord1,
			const Ordinal&ord2):
			IterFuncOrdinal(*new AdmisLevOrdinalImpl(indexCK,ord1,
				createParameters(&ord2))){}
				
		AdmisLevOrdinal (const Ordinal& indexCK, const Ordinal& ord1,
			const Ordinal&ord2, const Ordinal& ord3):IterFuncOrdinal(
			*new AdmisLevOrdinalImpl(indexCK,
			ord1,createParameters(&ord2,&ord3))){}

		AdmisLevOrdinal (const Ordinal& indexCK, const Ordinal& ord1,
			const Ordinal&ord2,
			const Ordinal& ord3, const Ordinal&ord4):IterFuncOrdinal(
			*new AdmisLevOrdinalImpl(indexCK,ord1,
			createParameters(&ord2,&ord3,&ord4))){}

		AdmisLevOrdinal (const Ordinal& indexCK, const Ordinal& iter,
			const Ordinal* const * const params=NULL,
			const Ordinal& drillDown = Ordinal::zero,
            const Embeddings& embed=Embeddings::embedNone):
			    IterFuncOrdinal(*new AdmisLevOrdinalImpl(indexCK,iter,params,
				    drillDown,embed)){}

		AdmisLevOrdinal(const AdmisLevOrdinalImpl& ord):
			IterFuncOrdinal(ord){}

	
		virtual ~AdmisLevOrdinal(){}
		
		
		enum {indexCKmaxParam = -2};
		static bool fixedPoint(const OrdinalImpl& ixCK, const Ordinal& iter,
            int ix, const Ordinal* const * const params, const Embeddings& e);

		static void texDocument() ;


	};




	


}
